@@ -19,9 +19,12 @@
 // OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
 // USE OR OTHER DEALINGS IN THE SOFTWARE.
 
+var StringDecoder = require('string_decoder').StringDecoder;
 var EventEmitter = require('events').EventEmitter;
 var net = require('net');
+var dgram = require('dgram');
 var Process = process.binding('process_wrap').Process;
+var assert = require('assert');
 var util = require('util');
 var constants; // if (!constants) constants = process.binding('constants');
 
@@ -65,7 +68,6 @@ function createSocket(pipe, readable) {
   if (readable) {
     s.writable = false;
     s.readable = true;
-    s.resume();
   } else {
     s.writable = true;
     s.readable = false;
@@ -109,29 +111,25 @@ var handleConversion = {
 
   'net.Socket': {
     send: function(message, socket) {
-      // pause socket so no data is lost, will be resumed later
-      socket.pause();
-
-      // if the socket wsa created by net.Server
+      // if the socket was created by net.Server
       if (socket.server) {
         // the slave should keep track of the socket
         message.key = socket.server._connectionKey;
 
         var firstTime = !this._channel.sockets.send[message.key];
-
-        // add socket to connections list
         var socketList = getSocketList('send', this, message.key);
-        socketList.add(socket);
 
         // the server should no longer expose a .connection property
-        // and when asked to close it should query the socket status from slaves
-        if (firstTime) {
-          socket.server._setupSlave(socketList);
-        }
+        // and when asked to close it should query the socket status from
+        // the slaves
+        if (firstTime) socket.server._setupSlave(socketList);
+
+        // Act like socket is detached
+        socket.server._connections--;
       }
 
       // remove handle from socket object, it will be closed when the socket
-      // has been send
+      // will be sent
       var handle = socket._handle;
       handle.onread = function() {};
       socket._handle = null;
@@ -139,21 +137,56 @@ var handleConversion = {
       return handle;
     },
 
+    postSend: function(handle) {
+      // Close the Socket handle after sending it
+      handle.close();
+    },
+
     got: function(message, handle, emit) {
       var socket = new net.Socket({handle: handle});
       socket.readable = socket.writable = true;
-      socket.pause();
 
       // if the socket was created by net.Server we will track the socket
       if (message.key) {
 
         // add socket to connections list
         var socketList = getSocketList('got', this, message.key);
-        socketList.add(socket);
+        socketList.add({
+          socket: socket
+        });
       }
 
       emit(socket);
-      socket.resume();
+    }
+  },
+
+  'dgram.Native': {
+    simultaneousAccepts: false,
+
+    send: function(message, handle) {
+      return handle;
+    },
+
+    got: function(message, handle, emit) {
+      emit(handle);
+    }
+  },
+
+  'dgram.Socket': {
+    simultaneousAccepts: false,
+
+    send: function(message, socket) {
+      message.dgramType = socket.type;
+
+      return socket._handle;
+    },
+
+    got: function(message, handle, emit) {
+      var socket = new dgram.Socket(message.dgramType);
+
+      socket.bind(handle, function() {
+        emit(socket);
+      });
     }
   }
 };
@@ -165,39 +198,47 @@ function SocketListSend(slave, key) {
   var self = this;
 
   this.key = key;
-  this.list = [];
   this.slave = slave;
-
-  slave.once('disconnect', function() {
-    self.flush();
-  });
-
-  this.slave.on('internalMessage', function(msg) {
-    if (msg.cmd !== 'NODE_SOCKET_CLOSED' || msg.key !== self.key) return;
-    self.flush();
-  });
 }
 util.inherits(SocketListSend, EventEmitter);
 
-SocketListSend.prototype.add = function(socket) {
-  this.list.push(socket);
+SocketListSend.prototype._request = function(msg, cmd, callback) {
+  var self = this;
+
+  if (!this.slave.connected) return onclose();
+  this.slave.send(msg);
+
+  function onclose() {
+    self.slave.removeListener('internalMessage', onreply);
+    callback(new Error('Slave closed before reply'));
+  };
+
+  function onreply(msg) {
+    if (!(msg.cmd === cmd && msg.key === self.key)) return;
+    self.slave.removeListener('disconnect', onclose);
+    self.slave.removeListener('internalMessage', onreply);
+
+    callback(null, msg);
+  };
+
+  this.slave.once('disconnect', onclose);
+  this.slave.on('internalMessage', onreply);
 };
 
-SocketListSend.prototype.flush = function() {
-  var list = this.list;
-  this.list = [];
-
-  list.forEach(function(socket) {
-    socket.destroy();
-  });
-};
-
-SocketListSend.prototype.update = function() {
-  if (this.slave.connected === false) return;
-
-  this.slave.send({
-    cmd: 'NODE_SOCKET_FETCH',
+SocketListSend.prototype.close = function close(callback) {
+  this._request({
+    cmd: 'NODE_SOCKET_NOTIFY_CLOSE',
     key: this.key
+  }, 'NODE_SOCKET_ALL_CLOSED', callback);
+};
+
+SocketListSend.prototype.getConnections = function getConnections(callback) {
+  this._request({
+    cmd: 'NODE_SOCKET_GET_COUNT',
+    key: this.key
+  }, 'NODE_SOCKET_COUNT', function(err, msg) {
+    if (err) return callback(err);
+    callback(null, msg.count);
   });
 };
 
@@ -207,45 +248,50 @@ function SocketListReceive(slave, key) {
 
   var self = this;
 
+  this.connections = 0;
   this.key = key;
-  this.list = [];
   this.slave = slave;
 
-  slave.on('internalMessage', function(msg) {
-    if (msg.cmd !== 'NODE_SOCKET_FETCH' || msg.key !== self.key) return;
+  function onempty() {
+    if (!self.slave.connected) return;
 
-    if (self.list.length === 0) {
-      self.flush();
-      return;
-    }
-
-    self.on('itemRemoved', function removeMe() {
-      if (self.list.length !== 0) return;
-      self.removeListener('itemRemoved', removeMe);
-      self.flush();
+    self.slave.send({
+      cmd: 'NODE_SOCKET_ALL_CLOSED',
+      key: self.key
     });
+  }
+
+  this.slave.on('internalMessage', function(msg) {
+    if (msg.key !== self.key) return;
+
+    if (msg.cmd === 'NODE_SOCKET_NOTIFY_CLOSE') {
+      // Already empty
+      if (self.connections === 0) return onempty();
+
+      // Wait for sockets to get closed
+      self.once('empty', onempty);
+    } else if (msg.cmd === 'NODE_SOCKET_GET_COUNT') {
+      if (!self.slave.connected) return;
+      self.slave.send({
+        cmd: 'NODE_SOCKET_COUNT',
+        key: self.key,
+        count: self.connections
+      });
+    }
   });
 }
 util.inherits(SocketListReceive, EventEmitter);
 
-SocketListReceive.prototype.flush = function() {
-  this.list = [];
-
-  if (this.slave.connected) {
-    this.slave.send({
-      cmd: 'NODE_SOCKET_CLOSED',
-      key: this.key
-    });
-  }
-};
-
-SocketListReceive.prototype.add = function(socket) {
+SocketListReceive.prototype.add = function(obj) {
   var self = this;
-  this.list.push(socket);
 
-  socket.on('close', function() {
-    self.list.splice(self.list.indexOf(socket), 1);
-    self.emit('itemRemoved');
+  this.connections++;
+
+  // Notify previous owner of socket about its state change
+  obj.socket.once('close', function() {
+    self.connections--;
+
+    if (self.connections === 0) self.emit('empty');
   });
 };
 
@@ -259,29 +305,29 @@ function getSocketList(type, slave, key) {
   return socketList;
 }
 
+var INTERNAL_PREFIX = 'NODE_';
 function handleMessage(target, message, handle) {
-  //Filter out internal messages
-  //if cmd property begin with "_NODE"
+  var eventName = 'message';
   if (message !== null &&
       typeof message === 'object' &&
       typeof message.cmd === 'string' &&
-      message.cmd.indexOf('NODE_') === 0) {
-    target.emit('internalMessage', message, handle);
-  }
-  //Non-internal message
-  else {
-    target.emit('message', message, handle);
+      message.cmd.length > INTERNAL_PREFIX.length &&
+      message.cmd.slice(0, INTERNAL_PREFIX.length) === INTERNAL_PREFIX) {
+    eventName = 'internalMessage';
   }
+  target.emit(eventName, message, handle);
 }
 
 function setupChannel(target, channel) {
   target._channel = channel;
+  target._handleQueue = null;
 
+  var decoder = new StringDecoder('utf8');
   var jsonBuffer = '';
   channel.buffering = false;
   channel.onread = function(pool, offset, length, recvHandle) {
     if (pool) {
-      jsonBuffer += pool.toString('ascii', offset, offset + length);
+      jsonBuffer += decoder.write(pool.slice(offset, offset + length));
 
       var i, start = 0;
 
@@ -290,7 +336,13 @@ function setupChannel(target, channel) {
         var json = jsonBuffer.slice(start, i);
         var message = JSON.parse(json);
 
-        handleMessage(target, message, recvHandle);
+        // There will be at most one NODE_HANDLE message in every chunk we
+        // read because SCM_RIGHTS messages don't get coalesced. Make sure
+        // that we deliver the handle with the right message however.
+        if (message && message.cmd === 'NODE_HANDLE')
+          handleMessage(target, message, recvHandle);
+        else
+          handleMessage(target, message, undefined);
 
         start = i + 1;
       }
@@ -311,8 +363,22 @@ function setupChannel(target, channel) {
 
   // handlers will go through this
   target.on('internalMessage', function(message, handle) {
+    // Once acknowledged - continue sending handles.
+    if (message.cmd === 'NODE_HANDLE_ACK') {
+      assert(Array.isArray(target._handleQueue));
+      var queue = target._handleQueue;
+      target._handleQueue = null;
+      queue.forEach(function(args) {
+        target.send(args.message, args.handle);
+      });
+      return;
+    }
+
     if (message.cmd !== 'NODE_HANDLE') return;
 
+    // Acknowledge handle receival.
+    target.send({ cmd: 'NODE_HANDLE_ACK' });
+
     var obj = handleConversion[message.type];
 
     // Update simultaneous accepts on Windows
@@ -342,18 +408,29 @@ function setupChannel(target, channel) {
       // this message will be handled by an internalMessage event handler
       message = {
         cmd: 'NODE_HANDLE',
-        type: 'net.',
+        type: null,
         msg: message
       };
 
-      switch (handle.constructor.name) {
-        case 'Socket':
-          message.type += 'Socket'; break;
-        case 'Server':
-          message.type += 'Server'; break;
-        case 'Pipe':
-        case 'TCP':
-          message.type += 'Native'; break;
+      if (handle instanceof net.Socket) {
+        message.type = 'net.Socket';
+      } else if (handle instanceof net.Server) {
+        message.type = 'net.Server';
+      } else if (handle instanceof process.binding('tcp_wrap').TCP ||
+                 handle instanceof process.binding('pipe_wrap').Pipe) {
+        message.type = 'net.Native';
+      } else if (handle instanceof dgram.Socket) {
+        message.type = 'dgram.Socket';
+      } else if (handle instanceof process.binding('udp_wrap').UDP) {
+        message.type = 'dgram.Native';
+      } else {
+        throw new TypeError("This handle type can't be sent");
+      }
+
+      // Queue-up message and handle if we haven't received ACK yet.
+      if (this._handleQueue) {
+        this._handleQueue.push({ message: message.msg, handle: handle });
+        return;
       }
 
       var obj = handleConversion[message.type];
@@ -365,22 +442,29 @@ function setupChannel(target, channel) {
       if (obj.simultaneousAccepts) {
         net._setSimultaneousAccepts(handle);
       }
+    } else if (this._handleQueue) {
+      // Queue request anyway to avoid out-of-order messages.
+      this._handleQueue.push({ message: message, handle: null });
+      return;
     }
 
     var string = JSON.stringify(message) + '\n';
     var writeReq = channel.writeUtf8String(string, handle);
 
-    // Close the Socket handle after sending it
-    if (message && message.type === 'net.Socket') {
-      handle.close();
-    }
-
     if (!writeReq) {
-      var er = errnoException(errno, 'write', 'cannot write to IPC channel.');
+      var er = errnoException(process._errno,
+                              'write',
+                              'cannot write to IPC channel.');
       this.emit('error', er);
+    } else if (handle && !this._handleQueue) {
+      this._handleQueue = [];
     }
 
-    writeReq.oncomplete = nop;
+    if (obj && obj.postSend) {
+      writeReq.oncomplete = obj.postSend.bind(null, handle);
+    } else {
+      writeReq.oncomplete = nop;
+    }
 
     /* If the master is > 2 read() calls behind, please stop sending. */
     return channel.writeQueueSize < (65536 * 2);
@@ -444,7 +528,9 @@ exports.fork = function(modulePath /*, args, options*/) {
   options.stdio = options.silent ? ['pipe', 'pipe', 'pipe', 'ipc'] :
       [0, 1, 2, 'ipc'];
 
-  return spawn(process.execPath, args, options);
+  options.execPath = options.execPath || process.execPath;
+
+  return spawn(options.execPath, args, options);
 };
 
 
@@ -452,7 +538,18 @@ exports._forkChild = function(fd) {
   // set process.send()
   var p = createPipe(true);
   p.open(fd);
+  p.unref();
   setupChannel(process, p);
+
+  var refs = 0;
+  process.on('newListener', function(name) {
+    if (name !== 'message' && name !== 'disconnect') return;
+    if (++refs === 1) p.ref();
+  });
+  process.on('removeListener', function(name) {
+    if (name !== 'message' && name !== 'disconnect') return;
+    if (--refs === 0) p.unref();
+  });
 };
 
 
@@ -578,7 +675,7 @@ exports.execFile = function(file /* args, options, callback */) {
   child.stdout.addListener('data', function(chunk) {
     stdout += chunk;
     if (stdout.length > options.maxBuffer) {
-      err = new Error('maxBuffer exceeded.');
+      err = new Error('stdout maxBuffer exceeded.');
       kill();
     }
   });
@@ -586,7 +683,7 @@ exports.execFile = function(file /* args, options, callback */) {
   child.stderr.addListener('data', function(chunk) {
     stderr += chunk;
     if (stderr.length > options.maxBuffer) {
-      err = new Error('maxBuffer exceeded.');
+      err = new Error('stderr maxBuffer exceeded.');
       kill();
     }
   });
@@ -643,10 +740,15 @@ function maybeClose(subprocess) {
 function ChildProcess() {
   EventEmitter.call(this);
 
+  // Initialize TCPWrap and PipeWrap
+  process.binding('tcp_wrap');
+  process.binding('pipe_wrap');
+
   var self = this;
 
   this._closesNeeded = 1;
   this._closesGot = 0;
+  this.connected = false;
 
   this.signalCode = null;
   this.exitCode = null;
@@ -662,6 +764,12 @@ function ChildProcess() {
     // - normally terminated processes don't touch this.signalCode
     // - signaled processes don't touch this.exitCode
     //
+    // new in 0.9.x:
+    //
+    // - spawn failures are reported with exitCode == -1
+    //
+    var err = (exitCode == -1) ? errnoException(process._errno, 'spawn') : null;
+
     if (signalCode) {
       self.signalCode = signalCode;
     } else {
@@ -675,7 +783,21 @@ function ChildProcess() {
     self._handle.close();
     self._handle = null;
 
-    self.emit('exit', self.exitCode, self.signalCode);
+    if (exitCode == -1) {
+      self.emit('error', err);
+    } else {
+      self.emit('exit', self.exitCode, self.signalCode);
+    }
+
+    // if any of the stdio streams have not been touched,
+    // then pull all the data through so that it can get the
+    // eof and emit a 'close' event.
+    // Do it on nextTick so that the user has one last chance
+    // to consume the output, if for example they only want to
+    // start reading the data once the process exits.
+    process.nextTick(function() {
+      flushStdio(self);
+    });
 
     maybeClose(self);
   };
@@ -683,6 +805,17 @@ function ChildProcess() {
 util.inherits(ChildProcess, EventEmitter);
 
 
+function flushStdio(subprocess) {
+  subprocess.stdio.forEach(function(stream, fd, stdio) {
+    if (!stream || !stream.readable || stream._consuming ||
+        stream._readableState.flowing)
+      return;
+    stream.resume();
+  });
+}
+
+
+
 function getHandleWrapType(stream) {
   if (stream instanceof handleWraps.Pipe) return 'pipe';
   if (stream instanceof handleWraps.TTY) return 'tty';
@@ -791,7 +924,7 @@ ChildProcess.prototype.spawn = function(options) {
 
     this._handle.close();
     this._handle = null;
-    throw errnoException(errno, 'spawn');
+    throw errnoException(process._errno, 'spawn');
   }
 
   this.pid = this._handle.pid;
@@ -807,9 +940,9 @@ ChildProcess.prototype.spawn = function(options) {
     if (stdio.handle) {
       // when i === 0 - we're dealing with stdin
       // (which is the only one writable pipe)
-      stdio.socket = createSocket(stdio.handle, i > 0);
+      stdio.socket = createSocket(self.pid !== 0 ? stdio.handle : null, i > 0);
 
-      if (i > 0) {
+      if (i > 0 && self.pid !== 0) {
         self._closesNeeded++;
         stdio.socket.on('close', function() {
           maybeClose(self);
@@ -876,14 +1009,14 @@ ChildProcess.prototype.kill = function(sig) {
       /* Success. */
       this.killed = true;
       return true;
-    } else if (errno == 'ESRCH') {
+    } else if (process._errno == 'ESRCH') {
       /* Already dead. */
-    } else if (errno == 'EINVAL' || errno == 'ENOSYS') {
+    } else if (process._errno == 'EINVAL' || process._errno == 'ENOSYS') {
       /* The underlying platform doesn't support this signal. */
-      throw errnoException(errno, 'kill');
+      throw errnoException(process._errno, 'kill');
     } else {
       /* Other error, almost certainly EPERM. */
-      this.emit('error', errnoException(errno, 'kill'));
+      this.emit('error', errnoException(process._errno, 'kill'));
     }
   }
 
