@@ -19,12 +19,18 @@
 // OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
 // USE OR OTHER DEALINGS IN THE SOFTWARE.
 
+var assert = require('assert');
 var util = require('util');
 var events = require('events');
 
 var UDP = process.binding('udp_wrap').UDP;
 
+var BIND_STATE_UNBOUND = 0;
+var BIND_STATE_BINDING = 1;
+var BIND_STATE_BOUND = 2;
+
 // lazily loaded
+var cluster = null;
 var dns = null;
 var net = null;
 
@@ -43,11 +49,6 @@ function isIP(address) {
 
 
 function lookup(address, family, callback) {
-  // implicit 'bind before send' needs to run on the same tick
-  var matchedFamily = isIP(address);
-  if (matchedFamily)
-    return callback(null, address, matchedFamily);
-
   if (!dns)
     dns = require('dns');
 
@@ -87,6 +88,24 @@ function newHandle(type) {
 }
 
 
+exports._createSocketHandle = function(address, port, addressType, fd) {
+  // Opening an existing fd is not supported for UDP handles.
+  assert(typeof fd !== 'number' || fd < 0);
+
+  var handle = newHandle(addressType);
+
+  if (port || address) {
+    var r = handle.bind(address, port || 0, 0);
+    if (r == -1) {
+      handle.close();
+      handle = null;
+    }
+  }
+
+  return handle;
+};
+
+
 function Socket(type, listener) {
   events.EventEmitter.call(this);
 
@@ -95,7 +114,7 @@ function Socket(type, listener) {
 
   this._handle = handle;
   this._receiving = false;
-  this._bound = false;
+  this._bindState = BIND_STATE_UNBOUND;
   this.type = type;
   this.fd = null; // compatibility hack
 
@@ -111,30 +130,87 @@ exports.createSocket = function(type, listener) {
 };
 
 
-Socket.prototype.bind = function(port, address) {
+function startListening(socket) {
+  socket._handle.onmessage = onMessage;
+  // Todo: handle errors
+  socket._handle.recvStart();
+  socket._receiving = true;
+  socket._bindState = BIND_STATE_BOUND;
+  socket.fd = -42; // compatibility hack
+
+  socket.emit('listening');
+}
+
+function replaceHandle(self, newHandle) {
+
+  // Set up the handle that we got from master.
+  newHandle.lookup = self._handle.lookup;
+  newHandle.bind = self._handle.bind;
+  newHandle.send = self._handle.send;
+  newHandle.owner = self;
+
+  // Replace the existing handle by the handle we got from master.
+  self._handle.close();
+  self._handle = newHandle;
+}
+
+Socket.prototype.bind = function(/*port, address, callback*/) {
   var self = this;
 
   self._healthCheck();
 
+  if (this._bindState != BIND_STATE_UNBOUND)
+    throw new Error('Socket is already bound');
+
+  this._bindState = BIND_STATE_BINDING;
+
+  if (typeof arguments[arguments.length - 1] === 'function')
+    self.once('listening', arguments[arguments.length - 1]);
+
+  var UDP = process.binding('udp_wrap').UDP;
+  if (arguments[0] instanceof UDP) {
+    replaceHandle(self, arguments[0]);
+    startListening(self);
+    return;
+  }
+
+  var port = arguments[0];
+  var address = arguments[1];
+  if (typeof address === 'function') address = '';  // a.k.a. "any address"
+
   // resolve address first
   self._handle.lookup(address, function(err, ip) {
-    if (!err) {
-      if (self._handle.bind(ip, port || 0, /*flags=*/0)) {
-        err = errnoException(errno, 'bind');
-      }
-      else {
-        self._bound = true;
-        self._startReceiving();
-        self.emit('listening');
-      }
+    if (err) {
+      self._bindState = BIND_STATE_UNBOUND;
+      self.emit('error', err);
+      return;
     }
 
-    if (err) {
-      // caller may not have had a chance yet to register its
-      // error event listener so defer the error to the next tick
-      process.nextTick(function() {
-        self.emit('error', err);
+    if (!cluster)
+      cluster = require('cluster');
+
+    if (cluster.isWorker) {
+      cluster._getServer(self, ip, port, self.type, -1, function(handle) {
+        if (!self._handle)
+          // handle has been closed in the mean time.
+          return handle.close();
+
+        replaceHandle(self, handle);
+        startListening(self);
       });
+
+    } else {
+      if (!self._handle)
+        return; // handle has been closed in the mean time
+
+      if (self._handle.bind(ip, port || 0, /*flags=*/ 0)) {
+        self.emit('error', errnoException(process._errno, 'bind'));
+        self._bindState = BIND_STATE_UNBOUND;
+        // Todo: close?
+        return;
+      }
+
+      startListening(self);
     }
   });
 };
@@ -168,16 +244,50 @@ Socket.prototype.send = function(buffer,
   if (!Buffer.isBuffer(buffer))
     throw new TypeError('First argument must be a buffer object.');
 
+  offset = offset | 0;
+  if (offset < 0)
+    throw new RangeError('Offset should be >= 0');
+
   if (offset >= buffer.length)
-    throw new Error('Offset into buffer too large');
+    throw new RangeError('Offset into buffer too large');
+
+  // Sending a zero-length datagram is kind of pointless but it _is_
+  // allowed, hence check that length >= 0 rather than > 0.
+  length = length | 0;
+  if (length < 0)
+    throw new RangeError('Length should be >= 0');
 
   if (offset + length > buffer.length)
-    throw new Error('Offset + length beyond buffer length');
+    throw new RangeError('Offset + length beyond buffer length');
+
+  port = port | 0;
+  if (port <= 0 || port > 65535)
+    throw new RangeError('Port should be > 0 and < 65536');
 
   callback = callback || noop;
 
   self._healthCheck();
-  self._startReceiving();
+
+  if (self._bindState == BIND_STATE_UNBOUND)
+    self.bind(0, null);
+
+  // If the socket hasn't been bound yet, push the outbound packet onto the
+  // send queue and send after binding is complete.
+  if (self._bindState != BIND_STATE_BOUND) {
+    // If the send queue hasn't been initialized yet, do it, and install an
+    // event handler that flushes the send queue after binding is done.
+    if (!self._sendQueue) {
+      self._sendQueue = [];
+      self.once('listening', function() {
+        // Flush the send queue.
+        for (var i = 0; i < self._sendQueue.length; i++)
+          self.send.apply(self, self._sendQueue[i]);
+        self._sendQueue = undefined;
+      });
+    }
+    self._sendQueue.push([buffer, offset, length, port, address, callback]);
+    return;
+  }
 
   self._handle.lookup(address, function(err, ip) {
     if (err) {
@@ -192,7 +302,7 @@ Socket.prototype.send = function(buffer,
       }
       else {
         // don't emit as error, dgram_legacy.js compatibility
-        var err = errnoException(errno, 'send');
+        var err = errnoException(process._errno, 'send');
         process.nextTick(function() {
           callback(err);
         });
@@ -205,10 +315,6 @@ Socket.prototype.send = function(buffer,
 function afterSend(status, handle, req, buffer) {
   var self = handle.owner;
 
-  // CHECKME socket's been closed by user, don't call callback?
-  if (handle !== self._handle)
-    void(0);
-
   if (req.cb)
     req.cb(null, buffer.length); // compatibility with dgram_legacy.js
 }
@@ -228,7 +334,7 @@ Socket.prototype.address = function() {
 
   var address = this._handle.getsockname();
   if (!address)
-    throw errnoException(errno, 'getsockname');
+    throw errnoException(process._errno, 'getsockname');
 
   return address;
 };
@@ -236,7 +342,7 @@ Socket.prototype.address = function() {
 
 Socket.prototype.setBroadcast = function(arg) {
   if (this._handle.setBroadcast((arg) ? 1 : 0)) {
-    throw errnoException(errno, 'setBroadcast');
+    throw errnoException(process._errno, 'setBroadcast');
   }
 };
 
@@ -247,7 +353,7 @@ Socket.prototype.setTTL = function(arg) {
   }
 
   if (this._handle.setTTL(arg)) {
-    throw errnoException(errno, 'setTTL');
+    throw errnoException(process._errno, 'setTTL');
   }
 
   return arg;
@@ -260,7 +366,7 @@ Socket.prototype.setMulticastTTL = function(arg) {
   }
 
   if (this._handle.setMulticastTTL(arg)) {
-    throw errnoException(errno, 'setMulticastTTL');
+    throw errnoException(process._errno, 'setMulticastTTL');
   }
 
   return arg;
@@ -271,7 +377,7 @@ Socket.prototype.setMulticastLoopback = function(arg) {
   arg = arg ? 1 : 0;
 
   if (this._handle.setMulticastLoopback(arg)) {
-    throw errnoException(errno, 'setMulticastLoopback');
+    throw errnoException(process._errno, 'setMulticastLoopback');
   }
 
   return arg; // 0.4 compatibility
@@ -287,7 +393,7 @@ Socket.prototype.addMembership = function(multicastAddress,
   }
 
   if (this._handle.addMembership(multicastAddress, interfaceAddress)) {
-    throw new errnoException(errno, 'addMembership');
+    throw new errnoException(process._errno, 'addMembership');
   }
 };
 
@@ -301,7 +407,7 @@ Socket.prototype.dropMembership = function(multicastAddress,
   }
 
   if (this._handle.dropMembership(multicastAddress, interfaceAddress)) {
-    throw new errnoException(errno, 'dropMembership');
+    throw new errnoException(process._errno, 'dropMembership');
   }
 };
 
@@ -312,34 +418,10 @@ Socket.prototype._healthCheck = function() {
 };
 
 
-Socket.prototype._startReceiving = function() {
-  if (this._receiving)
-    return;
-
-  if (!this._bound) {
-    this.bind(); // bind to random port
-
-    // sanity check
-    if (!this._bound)
-      throw new Error('implicit bind failed');
-  }
-
-  this._handle.onmessage = onMessage;
-  this._handle.recvStart();
-  this._receiving = true;
-  this.fd = -42; // compatibility hack
-};
-
-
 Socket.prototype._stopReceiving = function() {
   if (!this._receiving)
     return;
 
-  // Black hole messages coming in when reading is stopped. Libuv might do
-  // this, but node applications (e.g. test/simple/test-dgram-pingpong) may
-  // not expect it.
-  this._handle.onmessage = noop;
-
   this._handle.recvStop();
   this._receiving = false;
   this.fd = null; // compatibility hack
@@ -348,12 +430,25 @@ Socket.prototype._stopReceiving = function() {
 
 function onMessage(handle, slab, start, len, rinfo) {
   var self = handle.owner;
-  if (!slab) return self.emit('error', errnoException(errno, 'recvmsg'));
+  if (!slab) {
+    return self.emit('error', errnoException(process._errno, 'recvmsg'));
+  }
   rinfo.size = len; // compatibility
   self.emit('message', slab.slice(start, start + len), rinfo);
 }
 
 
+Socket.prototype.ref = function() {
+  if (this._handle)
+    this._handle.ref();
+};
+
+
+Socket.prototype.unref = function() {
+  if (this._handle)
+    this._handle.unref();
+};
+
 // TODO share with net_uv and others
 function errnoException(errorno, syscall) {
   var e = new Error(syscall + ' ' + errorno);
